@@ -3,21 +3,31 @@
 #
 # + pack/unpack everywhere
 # + db-sorted iteritems
-# - inform about broken ordering of anything but strings
+# |- inform about broken ordering of anything but strings
 # + ns functionality transparently optional
-# - create base-class with a mapping of an existing table 
-# - move ns- and flags stuff to dedicated functions
+# + create class with a mapping of an existing table
+# + move ns- and flags stuff to dedicated functions
+#
+# - add user controlled key preprocessor and val processor
+# - add raw query template as valgetter
+# - if valgetter is a string, it's used as an SQL template
+#       which is fed with key to obtain val
+# - build bunch of rawsql-dicts and build code snipplets
+#       with which to extract values
+# - build programs using these rawsql-dicts as components
+# - joins work just as well, as it's raw dicts
+# - iterations simply leave out which constraint
+# keymod, valmod, valmodin alongside valgetter
 #
 #
 __all__ = ['sqldict']
-
 __doc__ = '''
-
     sqldict - dict with sqlalchemy database-agnostic back-end
               capable of using picklable objects as both
               keys and values
 '''
 
+
 import doctest
 from UserDict import DictMixin
 import pickle
@@ -43,7 +53,32 @@ _KEYSIZE = 1000-64
 _VALSIZE = 4000
 _NSSIZE  = 64
 
-class SQLDict(DictMixin):
+_DEFAULT_KEYCOL   = "sqldict_key"
+_DEFAULT_VALCOL   = "sqldict_val"
+_DEFAULT_NSCOL    = "sqldict_ns"
+_DEFAULT_FLAGSCOL = "sqldict_flags"
+_DEFAULT_VALGETTER = lambda row:row[_DEFAULT_VALCOL]
+_DEFAULT_KEYMOD   = lambda key:key
+_DEFAULT_VALMODIN = lambda val:val
+_DEFAULT_VALMODOUT = lambda val:val
+
+_DEFAULT_KEYTYPE=String(_KEYSIZE)
+_DEFAULT_VALTYPE=String(_VALSIZE)
+
+
+def log_execute(conn, query):
+    #print str(query)
+    result = conn.execute(query)
+    return result
+
+
+class SetConfig(object):
+    def set(self, **kw):
+        for key, val in kw.iteritems():
+            setattr(self, "_"+key, val)
+
+
+class SQLDict(DictMixin, SetConfig):
     '''
     >>> e = create_engine('sqlite://')
     >>> d = SQLDict(e.connect(), 'asd', create=1)
@@ -59,15 +94,20 @@ class SQLDict(DictMixin):
     def __init__(self,  conn, tablename,
                             create=0,
                             ns='',
-                            keycol='sqldict_key',
-                            valcol='sqldict_val',
-                            keytype=String(_KEYSIZE),
-                            valtype=String(_VALSIZE),
-                            val_getter=lambda row:row['sqldict_val'],
+                            keycol=_DEFAULT_KEYCOL,
+                            valcol=_DEFAULT_VALCOL,
+                            nscol=_DEFAULT_NSCOL,
+                            flagscol=_DEFAULT_FLAGSCOL,
+                            keytype=_DEFAULT_KEYTYPE,
+                            valtype=_DEFAULT_KEYTYPE,
+                            valgetter=_DEFAULT_VALGETTER,
+                            keymod=_DEFAULT_KEYMOD,
+                            valmodin=_DEFAULT_VALMODIN,
+                            valmodout=_DEFAULT_VALMODOUT,
                             extended=0, # 0 or 1; simple or complex types
                             sort=1,         # 0=dont, 1=keys, 2=values(broken)
-                            keysize=_KEYSIZE,
-                            valsize=_VALSIZE,
+                            iterate=0,
+
                             nssize=_NSSIZE,
                             ):
         '''
@@ -77,28 +117,37 @@ class SQLDict(DictMixin):
         else:
             self._conn = conn
         self._tablename = tablename
-        self._keycol = keycol
-        self._valcol = valcol
+        self._keycol = keycol #as string name
+        self._valcol = valcol #as string name
+        self._nscol = nscol #as string name
+        self._flagscol = flagscol
         self._keytype = keytype
         self._valtype = valtype
-        self._val_getter = val_getter
+        self._valgetter = valgetter
+        self._keymod = keymod
+        self._valmodin = valmodin
+        self._valmodout = valmodout
         self._extended = extended
         self._ns = ns
         self._create = create
         self._sort = sort
-        self._keysize = keysize
-        self._valsize =  valsize
+        self._iterate = iterate
         self._nssize =  nssize
         self._meta = MetaData()
         self._meta.bind = conn
         self._keycoltype = None # set by _autoload_table()
         self._keyvaltype = None # set by _autoload_table()
         #
-        if valcol != 'sqldict_val':
-            if 0 and valcol:
+
+        #
+        # If user changes valcol but not valgetter we need
+        # to set it for him, otherwise __getitem__ wont work.
+        #
+        if valcol != _DEFAULT_VALCOL:
+            if valgetter != _DEFAULT_VALGETTER:
                 pass
             else:
-                self._val_getter = lambda row:row[valcol]
+                self._valgetter = lambda row:row[valcol]
 
         #if create:
         #    if self._extended is None:
@@ -127,10 +176,10 @@ class SQLDict(DictMixin):
     def _create_table(self):
         keycol = self._keycol
         valcol = self._valcol
+        nscol = self._nscol
+        flagscol = self._flagscol
         keytype = self._keytype
         valtype = self._valtype
-        keysize = self._keysize
-        valsize = self._valsize
         nssize = self._nssize
         ns = self._ns
         extended = self._extended
@@ -141,11 +190,11 @@ class SQLDict(DictMixin):
         ]
         if extended:
             columns.append(
-                Column('sqldict_flags', Integer, nullable=False),
+                Column(flagscol, Integer, nullable=False),
             )
         if ns:
             columns.append(
-                Column('sqldict_ns',  String(nssize), nullable=False, primary_key=True)
+                Column(nscol,  String(nssize), nullable=False, primary_key=True)
             )
         table = Table(self._tablename, self._meta, *columns)
         self._table = table
@@ -167,6 +216,7 @@ class SQLDict(DictMixin):
         conn = self._conn
         keycol = self._keycol
         valcol = self._valcol
+        flagscol = self._flagscol
         extended = self._extended
         key_column = self._get_key_column()
         #
@@ -174,16 +224,15 @@ class SQLDict(DictMixin):
         query = self._apply_ns_where_clause(query)
         query = query.where(key_column == key)
 
-
         values = {
             valcol : val
         }
-        if extended:
-            values["sqldict_flags"] = flags
 
+        if extended:
+            values[flagscol] = flags
 
         query = query.values(**values)
-        conn.execute(query)
+        log_execute(conn, query)
 
     def drop(self):
         self._table.drop()
@@ -243,31 +292,35 @@ class SQLDict(DictMixin):
 
     def _get_key_column(self):
         table = self._table
-        key_column = getattr(table.c, self._keycol)
-        return key_column
+        return getattr(table.c, self._keycol)
 
     def _get_val_column(self):
         table = self._table
-        val_column = getattr(table.c, self._valcol)
-        return val_column
+        return getattr(table.c, self._valcol)
 
-    def _get_sort_column(self, flags):
+    def _get_ns_column(self):
         table = self._table
-        sort = self._sort
-        col = table.c.sqldict_val
-        if sort == 1:
-            if flags & self._F_VAL_INT:
-                pass
-        if sort == 2:
-            if flags & self._F_KEY_INT:
-                pass
-        return col
+        return getattr(table.c, self._nscol)
+
+    def _get_flags_column(self):
+        table = self._table
+        return getattr(table.c, self._flagscol)
+
+    #def _get_sort_column(self, flags):
+    #    "NOT USED"
+    #    sort = self._sort
+    #    #
+    #    if sort == 1:
+    #        return self._get_key_column()
+    #    elif sort == 2:
+    #        return self._get_val_column()
 
     def _apply_ns_where_clause(self, query):
         table = self._table
         ns = self._ns
         if ns:
-            query = query.where(table.c.sqldict_ns == ns)
+            ns_column = self._get_ns_column()
+            query = query.where(ns_column == ns)
         return query
 
     def _apply_key_where_clause(self, query, key):
@@ -280,10 +333,10 @@ class SQLDict(DictMixin):
         #
         if sort == 1:
             key_column = self._get_key_column()
-            query = query.order_by(key_column)
+            query = query.order_by( key_column)
         if sort == 2:
             val_column = self._get_val_column()
-            query = query.order_by(val_column)
+            query = query.order_by( val_column)
         return query
 
     def _select(self, columns):
@@ -291,7 +344,7 @@ class SQLDict(DictMixin):
         extended = self._extended
         #
         if extended:
-            columns += [table.c.sqldict_flags]
+            columns += [self._get_flags_column()]
         query = select(columns)
         query = self._apply_ns_where_clause(query)
         query = self._apply_ordering(query)
@@ -334,18 +387,19 @@ class SQLDict(DictMixin):
     def iteritems(self):
         conn = self._conn
         keycol = self._keycol
+        flagscol = self._flagscol
         extended = self._extended
-        val_getter = self._val_getter
+        valgetter = self._valgetter
         #
         query = self._select_items()
         #print query
-        result = conn.execute(query)
+        result = log_execute(conn, query)
         for row in result:
             key = row[keycol]
-            val = val_getter(row)
+            val = valgetter(row)
             flags = 0
             if extended:
-                flags = row['sqldict_flags']
+                flags = row[flagscol]
             key = self._unpack_key(key, flags)
             val = self._unpack_val(val, flags)
             yield (key, val)
@@ -353,52 +407,57 @@ class SQLDict(DictMixin):
     def iterkeys(self):
         conn = self._conn
         keycol = self._keycol
+        flagscol = self._flagscol
         extended = self._extended
         #
         query = self._select_keys()
-        result = conn.execute(query)
+        result = log_execute(conn, query)
         for row in result:
             key = row[keycol]
             flags = 0
             if extended:
-                flags = row['sqldict_flags']
+                flags = row[flagscol]
             key = self._unpack_key(key, flags)
             yield key
 
     def itervalues(self):
         conn = self._conn
         extended = self._extended
-        val_getter = self._val_getter
+        flagscol = self._flagscol
+        valgetter = self._valgetter
         #
         query = self._select_values()
-        result = conn.execute(query)
+        result = log_execute(conn, query)
         for row in result:
-            val = val_getter(row)
+            val = valgetter(row)
             flags = 0
             if extended:
-                flags = row['sqldict_flags']
+                flags = row[flagscol]
             val = self._unpack_val(val, flags)
             yield val
 
     def namespaces(self):
         conn = self._conn
         ns = self._ns
+        nscol = self._nscol
+        ns_column = None # set after `if not ns:`
         table = self._table
         sort = self._sort
         #
         if not ns:
             error = 'namespaces(): not supported, sqldict instantiated without `ns=`'
             raise Exception(error)
-        query = select([ distinct(table.c.sqldict_ns) ])
+        ns_column = self._get_ns_column()
+        query = select([ distinct(ns_column) ])
         if sort == 1:
-            query = query.order_by(table.c.sqldict_ns)
+            query = query.order_by(ns_column)
         if sort == 2:
-            query = query.order_by(table.c.sqldict_ns)
+            query = query.order_by(ns_column)
         #print query
-        result = conn.execute(query)
+        result = log_execute(conn, query)
         namespaces = []
         for row in result:
-            ns = row['sqldict_ns']
+            ns = row[nscol]
             namespaces.append(ns)
         return namespaces
 
@@ -409,8 +468,6 @@ class SQLDict(DictMixin):
             ns=self._ns,
             create=0,
             sort=self._sort,
-            keysize=self._keysize,
-            valsize=self._valsize,
             nssize=self._nssize,
         )
         return kwargs
@@ -428,64 +485,84 @@ class SQLDict(DictMixin):
         duplicate = cls(**kwargs)
         return duplicate
 
-        
     def keys(self):
         conn = self._conn
         extended = self._extended
+        keycol = self._keycol
+        flagscol = self._flagscol
         #
         query = self._select_keys()
         #print query
-        result = conn.execute(query)
+        result = log_execute(conn, query)
         keys = []
         for row in result:
-            key = row['sqldict_key']
+            key = row[keycol]
             flags = 0
             if extended:
-                flags = row['sqldict_flags']
+                flags = row[flagscol]
             key = self._unpack_key(key, flags)
             keys.append(key)
         return keys
 
-    def __getitem__(self, key):
+    def __iter_getitem__(self, key, ordering=0):
         #print 'getitem', key
         #if not type(key) in (str, unicode):
         #    key = self._serialize(key)
         conn = self._conn
         table = self._table
+        flagscol = self._flagscol
         key_column = self._get_key_column()
         extended = self._extended
+        valgetter = self._valgetter
+        keymod = self._keymod
+        valmodout = self._valmodout
         #
+        if keymod:
+            key = keymod(key)
         key, _ = self._pack_key(key, 0)
         query = select([table])
         query = self._apply_ns_where_clause(query)
         query = query.where( key_column == key)
-        if key == 'aper':
-            print query
+        if ordering:
+            query = self._apply_ordering(query)
 
-        result = conn.execute(query)
+        result = log_execute(conn, query)
         for row in result:
-            val_getter = self._val_getter
             #
-            val = val_getter(row)
+            val = valgetter(row)
             flags = 0
             if extended:
-                flags = row['sqldict_flags']
+                flags = row[flagscol]
+
             val = self._unpack_val(val, flags)
+            if valmodout:
+                val = valmodout(val)
+            yield val
+
+    def __getitem__(self, key):
+        iterate = self._iterate
+        if iterate:
+            return self.__iter_getitem__(key, ordering=1)
+        for val in self.__iter_getitem__(key):
             return val
         raise KeyError
 
+
     def __delitem__(self, key):
         #print 'delitem', key
         conn = self._conn
         table = self._table
+        keymod = self._keymod
         #
+        if keymod:
+            key = keymod(key)
         if not type(key) in (str, unicode):
             key = self._serialize(key)
         query = table.delete()
         query = self._apply_ns_where_clause(query)
         query = self._apply_key_where_clause(query, key)
 
-        result = conn.execute(query)
+        result = log_execute(conn, query)
         if result.rowcount == 0:
             raise KeyError
         elif result.rowcount > 1:
@@ -496,9 +573,18 @@ class SQLDict(DictMixin):
     def __setitem__(self, key, val):
         keycol = self._keycol
         valcol = self._valcol
+        flagscol = self._flagscol
+        nscol = self._nscol
         extended = self._extended
         ns = self._ns
+        keymod = self._keymod
+        valmodin = self._valmodin
         #
+        if keymod:
+            key = keymod(key)
+        if valmodin:
+            val = valmodin(val)
+
         key, flags = self._pack_key(key, 0)
         val, flags = self._pack_val(val, flags)
 
@@ -506,18 +592,18 @@ class SQLDict(DictMixin):
         table = self._table
         query = insert(table)
 
-        kwargs = {
+        values = {
             keycol : key,
             valcol : val,
         }
         if extended:
-            kwargs['sqldict_flags'] = flags
+            values[flagscol] = flags
         if ns:
-            kwargs['sqldict_ns'] = ns
+            values[nscol] = ns
 
-        query = query.values(**kwargs)
+        query = query.values(**values)
         try:
-            conn.execute(query)
+            log_execute(conn, query)
         except IntegrityError:
             self._update(key, val, flags)
 
@@ -573,15 +659,20 @@ del __engine
 
 
 def sqldict(db, tablename,
-                    ns=None,
-                    serialize=0,
-                    create=0,
-                    sort=1,
-                    keycol="sqldict_key",
-                    valcol="sqldict_val",
-                    keytype=String(_KEYSIZE),
-                    valtype=String(_VALSIZE)
-                    ):
+                    keycol=_DEFAULT_KEYCOL,
+                    valcol=_DEFAULT_VALCOL,
+                    keytype=_DEFAULT_KEYTYPE,
+                    valtype=_DEFAULT_VALTYPE,
+                    valgetter=_DEFAULT_VALGETTER,
+                    keymod=_DEFAULT_KEYMOD,
+                    valmodin=_DEFAULT_VALMODIN,
+                    valmodout=_DEFAULT_VALMODOUT,
+                        iterate=0,
+                        serialize=0,
+                        create=0,
+                        sort=1,
+                        ns=None,
+                        ):
 
     """
     The various dictionaries, dN, can be used just as an ordinary dict.
@@ -598,10 +689,13 @@ def sqldict(db, tablename,
 
     >>> _ = e.execute("create table asd (i integer, s varchar(50))")
     >>> _ = e.execute("insert into asd values (42, 'gubbe')")
+    >>> _ = e.execute("insert into asd values (99, 'hatt')")
     >>> d3 = sqldict(e, "asd", keycol="s", valcol="i")
     >>> d4 = sqldict(e, "asd", keycol="i", valcol="s")
+    >>> #d5 = sqldict(e, "asd", keycol="s", valgetter=lambda r:r)
     >>> assert d3["gubbe"] == 42
     >>> assert d4[42] == "gubbe"
+    >>> #assert d5["gubbe"]["i"] == 42
 
     broken (at least mysql) when combining keytype and serialize,
     it tries to serialize 42.
@@ -632,11 +726,16 @@ def sqldict(db, tablename,
                     ns=ns,
                     extended=serialize,
                     create=create,
-                    sort=1,
+                    sort=sort,
                     keycol=keycol,
                     valcol=valcol,
                     keytype=keytype,
                     valtype=valtype,
+                    valgetter=valgetter,
+                    keymod=keymod,
+                    valmodin=valmodin,
+                    valmodout=valmodout,
+                    iterate=iterate,
                     )
 
 
