@@ -1,4 +1,4 @@
-from sqldict import sqldict, sqldict_old
+from sqldict import sqldict, sqldict_old, _KEYSIZE, _VALSIZE
 
 import doctest
 import unittest
@@ -9,26 +9,60 @@ SQLITEFS_FILE = "//tmp/.sqlitedb_test_sqlitefs"
 
 def randstring(len):
     """
-    >>> assert 1==2
+    >>> assert 1==1
     """
     import uuid
     s = uuid.uuid4().__str__().replace("-", "")[:len]
     return s
 
 
-
 class Test_sqldict_basic(unittest.TestCase):
 
+    _mysql_uri = 'mysql://sqldict:balalajkaperestrojka@localhost/sqldict'
+
+    _engines = engines = {
+        "sqlite"   : create_engine("sqlite://"),
+        "sqlitefs" : create_engine("sqlite://"+SQLITEFS_FILE),
+        "mysql"    : create_engine(_mysql_uri),
+    }
+    _testtables = dict(
+        t_char_char = """
+            CREATE TABLE t_char_char (
+                xkey varchar(%(keysize)s),
+                xval varchar(%(valsize)s)
+            )
+        """,
+        t_int_char = """
+            CREATE TABLE t_int_char (
+                xkey integer,
+                xval varchar(%(valsize)s)
+            )
+        """,
+        t_int_int = """
+            CREATE TABLE t_int_int (
+                xkey integer,
+                xval integer
+            )
+        """,
+    )
+    def tablenames(self):
+        return self._testtables.iterkeys()
+
     def setUp(self):
-        mysql_uri = 'mysql://sqldict:balalajkaperestrojka@localhost/sqldict'
-        self._engines = engines = {
-            "sqlite"   : create_engine("sqlite://"),
-            "sqlitefs" : create_engine("sqlite://"+SQLITEFS_FILE),
-            "mysql"    : create_engine(mysql_uri),
-        }
+        for query_template in self._testtables.itervalues():
+            query = query_template % dict(
+                        keysize = _KEYSIZE,
+                        valsize = _VALSIZE,
+            )
+            for engine in self._engines.itervalues():
+                engine.execute(query)
 
     def tearDown(self):
         import os
+        for tablename in self._testtables.iterkeys():
+                query = "drop table "+ tablename
+                for engine in self._engines.itervalues():
+                    engine.execute(query)
         try:
             os.unlink(SQLITEFS_FILE)
         except:
@@ -58,13 +92,15 @@ class Test_sqldict_basic(unittest.TestCase):
         d.drop()
 
     def test_main(self):
-        assert len(self._engines) == 3
         engines = self._engines
         #
         for name, engine in self._engines.iteritems():
             self.compare( engine, self.set_get_del, "set_get_del")
             self.sorting(engine)
             self.serialization_and_types(engine)
+            self.custom_valgetter(engine)
+            self.keymod_valmod(engine)
+            self.iterate(engine)
             factor = 1
             self.large_update(engine, count=factor*10)
 
@@ -84,11 +120,11 @@ class Test_sqldict_basic(unittest.TestCase):
             d[rand] = "%d" % i
             ref[rand] = "%d" % i
         d._sort = 1
-        assert d.keys() == sorted(d.keys()) == sorted(ref.keys())
+        assert len(d.keys())   and d.keys()   == sorted(ref.keys())
         d._sort = 2
-        assert d.values() == sorted(d.values()) == sorted(ref.values())
+        assert len(d.values()) and d.values() == sorted(ref.values())
         d._sort = 1
-        assert d.items() == sorted(d.items()) == sorted(ref.items())
+        assert len(d.items())  and d.items()  == sorted(ref.items())
         d.drop()
 
     def large_update(self, engine, count=100):
@@ -101,6 +137,95 @@ class Test_sqldict_basic(unittest.TestCase):
         assert d == src
         d.drop()
 
+    def custom_valgetter(self, engine):
+        def get_row(row):
+            return row
+        d = sqldict(engine, "test_valgetter",
+                        keycol="kk",
+                        valcol="vv",
+                        valgetter=get_row,
+                        create=1,
+                        serialize=1,
+                        )
+        try:
+            d["asd"] = "qwe"
+            assert d["asd"]["kk"] == "asd"
+            assert d["asd"]["vv"] == "qwe"
+            assert d["asd"].kk == "asd"
+            assert d["asd"].vv == "qwe"
+        finally:
+            d.drop()
+
+    def keymod_valmod(self, engine):
+        def keymod(key):
+            return key+"KK"
+        def valmodout(val):
+            return val+"VV"
+        d = sqldict(engine, "test_valgetter",
+                        keycol="kk",
+                        valcol="vv",
+                        create=1,
+                        serialize=1,
+                        keymod=keymod,
+                        valmodout=valmodout,
+                        )
+        d2 = sqldict(engine, "test_valgetter",
+                        keycol="kk",
+                        valcol="vv",
+                        #create=1,
+                        serialize=1,
+                        #keymod=keymod,
+                        #valmodout=valmodout,
+                        )
+        try:
+            d["asd"] = "qwe"
+            assert d["asd"] == "qweVV"
+            assert d2["asdKK"] == "qwe"
+
+            d2.set(valmodin  = lambda val : val+val)
+            d2.set(valmodout = lambda val : val.upper())
+            d2["x"] = "y"
+            assert d2["x"] == "YY"
+        finally:
+            d.drop()
+
+    def iterate(self, engine):
+        for val in ["asd", "qwe", "123", "haha", "hoho"]:
+            engine.execute("""
+                INSERT INTO t_char_char
+                VALUES ('foo', '%s')
+            """ % val
+            )
+        d = sqldict(engine, "t_char_char",
+                        keycol="xkey",
+                        valcol="xval",
+                        iterate=1,
+                        sort=1,
+                        )
+        #
+        # sort by key
+        #
+        i = 0
+        for i, val in enumerate(d["foo"]):
+            if i==0: assert val == "asd"
+            if i==1: assert val == "qwe"
+            if i==2: assert val == "123"
+            if i==3: assert val == "haha"
+            if i==4: assert val == "hoho"
+        assert i==4
+        #
+        # sort by val
+        #
+        d.set(sort=2)
+        i = 0
+        for i, val in enumerate(d["foo"]):
+            if i==0: assert val == "123"
+            if i==1: assert val == "asd"
+            if i==2: assert val == "haha"
+            if i==3: assert val == "hoho"
+            if i==4: assert val == "qwe"
+        assert i==4
+
 
 
 
@@ -376,5 +501,5 @@ class Test_sqldict_old(unittest.TestCase):
 
 
 if __name__ == "__main__":
-    #doctest.testmod()
+    doctest.testmod()
     unittest.main()
